/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * This is GNU GO, a Go program. Contact gnugo@gnu.org, or see   *
 * http://www.gnu.org/software/gnugo/ for more information.      *
 *                                                               *
 * Copyright 1999 and 2000 by the Free Software Foundation.      *
 *                                                               *
 * This program is free software; you can redistribute it and/or *
 * modify it under the terms of the GNU General Public License   *
 * as published by the Free Software Foundation - version 2.     *
 *                                                               *
 * This program is distributed in the hope that it will be       *
 * useful, but WITHOUT ANY WARRANTY; without even the implied    *
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR       *
 * PURPOSE.  See the GNU General Public License in file COPYING  *
 * for more details.                                             *
 *                                                               *
 * You should have received a copy of the GNU General Public     *
 * License along with this program; if not, write to the Free    *
 * Software Foundation, Inc., 59 Temple Place - Suite 330,       *
 * Boston, MA 02111, USA                                         *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */




#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <math.h>

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "liberty.h"

#include "sgftree.h"
#include "analyze.h"

#define MAXTRY 400

/* Return one if x doesn't equal position_number and 0 otherwise.
 * After using this macro x will always have the value
 * position_number.
 */
#define NEEDS_UPDATE(x) (x != position_number ? (x = position_number, 1) : 0)

void sgfShowConsideredMoves(void);


/* Initialize the engine. This prepares the pattern matcher and
 * reading code for use.
 */

void
initialize_engine()
{
  static int node_limits[10] = {10000, 2000, 1000, 500, 300,
				200, 150, 100, 50, 20};
  /* Initialize things for hashing of positions. */
  reading_cache_clear();
  hashdata_recalc(&hashdata, p, ko_i, ko_j);

  /* Set up depth values. */
  depth               = max(6, nominal_depth - hurry);
  ko_depth            = max(1, nominal_ko_depth - hurry);
  backfill_depth      = max(2, nominal_backfill_depth - hurry);
  branch_depth        = max(3, nominal_branch_depth - hurry);
  fourlib_depth       = max(1, nominal_fourlib_depth - hurry + (hurry > 0));
  mouse_branch_depth  = max(2, nominal_mouse_branch_depth - hurry/2);
  mouse_reading_depth = max(5, nominal_mouse_reading_depth - hurry/2);

  if (hurry < 0)
    mouse_node_limit  = nominal_mouse_node_limit * pow(1.5, -hurry);
  else if (hurry < 10)
    mouse_node_limit  = (nominal_mouse_node_limit * node_limits[hurry] /
			 node_limits[0]);
  else
    mouse_node_limit  = (nominal_mouse_node_limit * node_limits[9] /
			 node_limits[0]);
  mouse_node_limit    = max(20, mouse_node_limit);
}


/*
 * Examine the position and try to gather as much information as possible.
 * This is used mainly for move generation, but could also be called
 * for debugging purposes (decidestring, etc).
 *
 * The parameter how_much tells us how much of the work we have to do.
 * For move generation we have to do it all.  For debugging we can 
 * sometimes stop a little earlier.
 */

void
examine_position(int color, int how_much)
{
  int  save_verbose;
  int  m, n;

  /* Position numbers for which various examinations were last made. */
  static int worms_examined = -1;
  static int initial_influence_examined = -1;
  static int dragons_examined = -1;
  static int mouse_reasons_examined = -1;
  static int moyo_examined = -1;
  static int initial_influence2_examined = -1;
  
  /* Don't print reading traces during make_worms and make_dragons unless 
   * the user really wants it (verbose == 3). 
   *
   * FIXME: We don't reset verbose to its correct value if we return within
   *        the if block.
   */
  save_verbose=verbose;
  if ((verbose==1) || (verbose==2))
    --verbose;

  if (NEEDS_UPDATE(worms_examined))
    make_worms();
  if (how_much == EXAMINE_WORMS)
    return;

  if (stones_on_board(BLACK | WHITE) != 0) {
    if (NEEDS_UPDATE(initial_influence_examined))
      compute_initial_influence(color, 0);
    if (how_much == EXAMINE_INITIAL_INFLUENCE)
      return;

    if (NEEDS_UPDATE(dragons_examined))
      make_dragons(color);
    if (how_much == EXAMINE_DRAGONS)
      return;

    if (NEEDS_UPDATE(mouse_reasons_examined))
      mouse_reasons(color);
    if (how_much == EXAMINE_MOUSE_REASONS)
      return;
  }
  else if (how_much == EXAMINE_INITIAL_INFLUENCE
	   || how_much == EXAMINE_DRAGONS
	   || how_much == EXAMINE_MOUSE_REASONS)
    return;
  verbose=save_verbose;

  /* Map the moyos for both sides.
   * FIXME: Probably the make_moyo() call should go into
   * make_dragons() at some point.
   */
  if (NEEDS_UPDATE(moyo_examined))
    make_moyo(color);
  if (how_much == EXAMINE_MOYO)
    return;

  if (printworms) {
    show_dragons();
  }
  else if (verbose) {
    for (m=0;m<board_size;m++)
      for (n=0;n<board_size;n++)
	if ((p[m][n])
	    && (worm[m][n].origini==m) 
	    && (worm[m][n].originj==n) 
	    && (worm[m][n].cutstone)) 
        {
	  TRACE("%scutting stones found at %m\n",
		worm[m][n].cutstone==2 ? "potential " : "", m,n); 
	}
  }	  

  if (NEEDS_UPDATE(initial_influence2_examined))
    compute_initial_influence(color, 1);
  if (how_much == EXAMINE_INITIAL_INFLUENCE2)
    return;
}


/* 
 * Generate computer move for COLOR.
 *
 * Return the generated move in (*i, *j).
 */
  
int
genmove(int *i, int *j, int color)
{
  float val;
  double t1=0., t2=0.;

  if (showtime)
    t1 = gg_gettimeofday();

  /* Prepare our table of moves considered. */
  memset(potential_moves, 0, sizeof(potential_moves));

  /* Prepare our table of move reasons. */
  clear_move_reasons();

  /* Reset all the statistics for each move. */
  stats.nodes = 0;
  stats.position_entered = 0;
  stats.position_hits    = 0;
  stats.read_result_entered = 0;
  stats.read_result_hits    = 0;
  stats.hash_collisions     = 0;

  /* no move is found yet. */
  *i = -1;  
  *j = -1;  
  val = -1; 

  /* Prepare pattern matcher and reading code. */
  initialize_engine();
  
  /* Find out information about the worms and dragons. */
  examine_position(color, EXAMINE_ALL);

  /*
   * Print some of the information if the user wants to.
   */
  if (printmoyo) {
/*    compute_influence();*/
    print_moyo(color);
  }
   
  if (printboard) {
    showboard(printboard); 
    if (printboard == 1)
      showboard(3);
  }

  assert(stackp == 0);

  /*
   * Ok, information gathering is complete. Now start to find some moves!
   */

  /* Try to find empty corner moves. */
  fuseki(color);
  assert(stackp == 0);

  /* pattern matcher */
  shapes(color);
  assert(stackp == 0);

  /* Review the move reasons and estimate move values. */
  if (review_move_reasons(i, j, &val, color)) {
    TRACE("Move generation likes %m with value %f\n", *i, *j, val);
    if (analyzerflag & ANALYZE_RECOMMENDED)
      sgfBoardText(0, *i, *j, "P");
  }
  assert(stackp == 0);
 
  /* If the move value is 6 or lower, we look for endgame patterns too. */
  if (val <= 6.0) {
    endgame_shapes(color);
    assert(stackp == 0);
    if (review_move_reasons(i, j, &val, color)) {
      TRACE("Move generation likes %m with value %f\n", *i, *j, val);
      if (analyzerflag & ANALYZE_RECOMMENDED)
	sgfBoardText(0, *i, *j, "P");
    }
    assert(stackp == 0);
  }
 
  
  /* If no move found yet, revisit any semeai and change the
   * status of the opponent group from DEAD to UNKNOWN, then 
   * run shapes and endgame_shapesagain. This may turn up a move.
   */
  if (val < 0.0) {
    if (revise_semeai(color)) {
      shapes(color);
      endgame_shapes(color);
      if (review_move_reasons(i, j, &val, color)) {
	TRACE("Upon reconsideration move generation likes %m with value %f\n",
	      *i, *j, val); 
	if (analyzerflag & ANALYZE_RECOMMENDED)
	  sgfBoardText(0,*i,*j,"p");
      }
    }
  }
  
  if ((val < 0.0) && fill_liberty(i, j, color)) {
    val = 1.0;
    TRACE("Filling a liberty at %m\n", *i, *j);
    if (analyzerflag & ANALYZE_RECOMMENDED)
      sgfBoardText(0, *i, *j, "L");
  }

  /* If no move is found then pass. */
  if (val < 0) {
    TRACE("I pass.\n");
    *i = -1;
    *j = -1;
  } else
    TRACE("genmove() recommends %m with value %f\n", *i, *j, val);
  
  /* If statistics is turned on, this is the place to show it. */
  if (showstatistics) {
    char text[120];
    sprintf(text, "Nodes:                %d\n", stats.nodes);
    sgfAddComment(0,text);
      
    printf("Nodes:                %d\n", stats.nodes);
    printf("Positions entered:    %d\n", stats.position_entered);
    printf("Position hits:        %d\n", stats.position_hits);
    printf("Read results entered: %d\n", stats.read_result_entered);
    printf("Read result hits:     %d\n", stats.read_result_hits);
    printf("Hash collisions:      %d\n", stats.hash_collisions);
  }

  if (showtime) {
    t2 = gg_gettimeofday();
    fprintf(stderr, "TIME to generate move: %f\n", t2-t1);
  }
  
  return val;
  
}  /* end genmove */




/* This is called for each move which has been considered. For
 * debugging purposes, we keep a table of all the moves we
 * have considered.
 */

void 
move_considered(int i, int j, int val)
{
  if ((val > potential_moves[i][j])) {
    potential_moves[i][j] = val;
  }
}



/*
 * Local Variables:
 * tab-width: 8
 * c-basic-offset: 2
 * End:
 */

