/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * This is GNU GO, a Go program. Contact gnugo@gnu.org, or see   *
 * http://www.gnu.org/software/gnugo/ for more information.      *
 *                                                               *
 * Copyright 1999 and 2000 by the Free Software Foundation.      *
 *                                                               *
 * This program is free software; you can redistribute it and/or *
 * modify it under the terms of the GNU General Public License   *
 * as published by the Free Software Foundation - version 2.     *
 *                                                               *
 * This program is distributed in the hope that it will be       *
 * useful, but WITHOUT ANY WARRANTY; without even the implied    *
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR       *
 * PURPOSE.  See the GNU General Public License in file COPYING  *
 * for more details.                                             *
 *                                                               *
 * You should have received a copy of the GNU General Public     *
 * License along with this program; if not, write to the Free    *
 * Software Foundation, Inc., 59 Temple Place - Suite 330,       *
 * Boston, MA 02111, USA                                         *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <assert.h>

#include "liberty.h"
#include "gg_utils.h"

/*
 * This function underpins all the TRACE and DEBUG stuff.
 * It accepts %c, %d, %f, %s, and %x as usual. But it
 * also accepts %m, which takes TWO integers and writes a move.
 * Other accepted formats are
 * %H: Print a hashvalue.
 * %C: Print a color as a string.
 * %M: Like %m but three characters wide for all moves (pads with spaces).
 * Nasty bodge: %o at the start means outdent, i.e. cancel indent.
 */

static void 
vgprintf(FILE* outputfile, const char *fmt, va_list ap)
{
  if (fmt[0] == '%' && fmt[1] == 'o')
    fmt += 2;  /* cancel indent */
  else if (stackp > 0)
    fprintf(outputfile, "%.*s", stackp*2, "                                ");

  for ( ; *fmt ; ++fmt ) {
    if (*fmt == '%') {
      switch (*++fmt) {
      case 'c':
      {
	/* rules of promotion => passed as int, not char */
	int c = va_arg(ap, int);
	putc(c, outputfile);
	break;
      }
      case 'd':
      {
	int d = va_arg(ap, int);
	fprintf(outputfile, "%d", d);
	break;
      }
      case 'x':
      {
	int d = va_arg(ap, int);
	fprintf(outputfile, "%x", d);
	break;
      }
      case 'f':
      {
	double f = va_arg(ap, double); /* passed as double, not float */
	fprintf(outputfile, "%f", f);
	break;
      }
      case 's':
      {
	char *s = va_arg(ap, char*);
	fputs(s, outputfile);
	break;
      }
      case 'm':
      case 'M':
      {
	char movename[4];
	int m = va_arg(ap, int);
	int n = va_arg(ap, int);
	if (m == -1 && n == -1)
	  fputs("PASS", outputfile);
	else if ((m<0) || (n<0) || (m>=board_size) || (n>=board_size))
	  fprintf(outputfile, "[%d,%d]", m, n);
	else {
	  /* Generate the move name. */
	  if (n<8)
	    movename[0] = n+65;
	  else
	    movename[0] = n+66;
	  if (*fmt == 'm')
	    sprintf(movename+1, "%d", board_size-m);
	  else
	    sprintf(movename+1, "%-2d", board_size-m);
	  fputs(movename, outputfile);
	}
	break;
      }
      case 'H':
      {
	long h = va_arg(ap, int);
	fprintf(outputfile, "%lx", h);
	break;
      }
      case 'C':
      {
	int color = va_arg(ap, int);
	fputs(color_to_string(color), outputfile);
	break;
      }
      default:
	fprintf(outputfile, "\n\nUnknown format character '%c'\n", *fmt);
	break;
      }
    }
    else
      putc(*fmt, outputfile);
  }
}


/*
 * required wrapper around vgprintf, writes to stderr
 */

void 
gprintf(const char *fmt, ...)
{
  va_list ap;
  va_start(ap, fmt);
  vgprintf(stderr, fmt, ap);
  va_end(ap);
}


/*
 * required wrapper around vgprintf, in contrast to gprintf this one
 * writes to stdout
 */

void
mprintf(const char *fmt, ...)
{
  va_list ap;
  va_start(ap, fmt);
  vgprintf(stdout, fmt, ap);
  va_end(ap);
}



#ifndef __GNUC__  /* gnuc allows us to define these as macros in liberty.h */

void 
TRACE(const char *fmt, ...)
{
  va_list ap;

  if (!verbose)
    return;

  va_start(ap, fmt);
  vprintf(fmt, ap);
  va_end(ap);
}


void 
VTRACE(const char *fmt, ...)
{
  va_list ap;

  if (verbose < 3)
    return;

  va_start(ap, fmt);
  vprintf(fmt, ap);
  va_end(ap);
}


void 
RTRACE(const char *fmt, ...)
{
  va_list ap;

  if (verbose < 2)
    return;

  va_start(ap, fmt);
  vprintf(fmt, ap);
  va_end(ap);
}


void 
DEBUG(int flag, const char *fmt, ...)
{
  va_list ap;

  if (!(debug & flag))
    return;

  va_start(ap, fmt);
  vprintf(fmt, ap);
  va_end(ap);
}

#endif


/*
 * A wrapper around abort() which shows the state variables at the time
 * of the problem.  (i, j) are typically a related move, or -1, -1.
 */

void 
abortgo(const char *file, int line, const char *msg, int x, int y)
{
  verbose = 4;
  TRACE("%o\n\n***assertion failure:\n%s:%d - %s near %m***\n\n",
	file, line, msg, x, y);
  dump_stack();
  showboard(0);

/* FIXME: See to it that we can dump the state again. */
#if 0
  while (stackp > 0) {
    popgo();
    showboard(0);
  }
#endif

#if 0
  if (sgf_root) {
    /* FIXME: How should we get seed and komi in here?
       writesgf(sgf_root, "abortgo.sgf", gameinfo->seed, gameinfo->komi); */
    sgf_write_header(sgf_root, analyzerflag & ANALYZE_OVERWRITE, 0, 0.0);
    writesgf(sgf_root, "abortgo.sgf");
  }
#endif

  fprintf(stderr, "\n\
gnugo %s: You stepped on a bug.\n\
Please save this game as an sgf file \
and mail together with the file abortgo.sgf to \
gnugo@gnu.org\n\n", gg_version());

  abort();  /* cause core dump */
}


/* Convert a color value to a string. */
const char *
color_to_string(int color)
{
  if (color == EMPTY)
    return "empty";
  else if (color == WHITE)
    return "white";
  else if (color == BLACK)
    return "black";
  else if (color == GRAY_BORDER)
    return "gray border";
  else if (color == WHITE_BORDER)
    return "white border";
  else if (color == BLACK_BORDER)
    return "black border";
  else
    return "purple?";
}

/* Convert a location to a string. */
const char *
location_to_string(int i, int j)
{
  static char buf[4];
  char *bufp = &buf[0];

  *bufp = 'A'+j;
  if (*bufp >= 'I')
    (*bufp)++;
  bufp++;

  i = board_size - i;
  if (i > 9)
    *bufp++ = '0' + i/10;
  *bufp++ = '0' + i%10;

  *bufp = 0;

  return buf;
}

/* Convert a status value to a string. */
const char *
status_to_string(int status)
{
  if (status == ALIVE)
    return "alive";
  else if (status == DEAD)
    return "dead";
  else if (status == UNKNOWN)
    return "unknown";
  else if (status == CRITICAL)
    return "critical";
  else if (status == UNCHECKED)
    return "unchecked";
  else
    return "ERROR";
}


/*
 * Local Variables:
 * tab-width: 8
 * c-basic-offset: 2
 * End:
 */
